
#include "alphabeta.h"
#include "connectivity.h"
#include "territory.h"
#include "time_util.h"

// size of transposition table in megabytes
#define TT_SIZE 300
// switch to endgame search when separated from opponent
#define COMBINED_SEARCH 1

Node::Node() : age(0), hash(0) { }

NodeTable::NodeTable()
{
    table_size = (TT_SIZE * 1024 * 1024) / sizeof(NodeTableCell);
    table = new NodeTableCell[table_size];
    current_age = 1;
}

const Node* NodeTable::get(const Map& position)
{
    Node* result;
    NodeTableCell* cell = &table[position.hash % table_size];
    if (cell->deepest.hash == position.hash)
    {
        result = &cell->deepest;
    }
    else
    {
        result = &cell->always;
    }
    return result;
}

void NodeTable::set(const Map& position, int ndepth,
        int nscore, Bound nbound, int norder)
{
    NodeTableCell* cell = &table[position.hash % table_size];
    Node* node;
    if (cell->deepest.age < current_age
            || cell->deepest.hash == position.hash
            || cell->deepest.depth < ndepth)
    {
        node = &cell->deepest;
    } else {
        node = &cell->always;
    }
    if (node->age == current_age && node->hash != position.hash)
        ++overwrite;
    node->age = current_age;
    node->hash = position.hash;
    node->depth = ndepth;
    node->score = nscore;
    node->bound = nbound;
    node->move_order = norder;
}

void NodeTable::ageTable()
{
    current_age += 1;
}

ABSearch::ABSearch(StopWatch* sw, NodeTable* table,
        cnct::TreeSizeCache* tsc, EndSearch* es)
: clock(sw), endgame(es)
{
    ttable = table;
    tscache = tsc;
}

int ABSearch::eval(Map& position,
        Territory* p1territory, Territory* p2territory)
{
    /* veronoi territory
    int p1score = p1territory->size(position.bikes[0]) * 10;
    int p2score = p2territory->size(position.bikes[1]) * 10;
    */
    Bound bound;
    int p1score = tscache->get_size(*p1territory, position.bikes[0],
            bound) * 10;
    int p2score = tscache->get_size(*p2territory, position.bikes[1],
            bound) * 10;
    int score = p1score - p2score;
    if (score >= MIN_WIN)
        score = MIN_WIN - 1;
    else if (score <= -MIN_WIN)
        score = -MIN_WIN + 1;
    return score;
}

int ABSearch::alphabeta(Map& position, int alpha, int beta, int depth)
{
    ++clock->nodes_visited;
    if (position.isEnd())
    {
        int score = position.score();
        if (score != 0)
        {
            return score * WIN;
        } else {
            return -1;
        }
    }
    int move_order = 0x3210;
    const Node* node = ttable->get(position);
    if (node->hash == position.hash)
    {
        if (node->depth >= depth
                &&(node->bound == EXACT
                    || (node->bound == LOWER && node->score <= alpha)
                    || (node->bound == UPPER && node->score >= beta)))
        {
            return node->score;
        }
        move_order = node->move_order;
    }
    Bound bound = LOWER;
    int score = -WIN;
    bool full_search = true;
    if (!position.moving)
    {
        Territory* p1territory = Territory::allocate();
        Territory* p2territory = Territory::allocate();
        bool shared_terr = position.space->separate(
            position.bikes[0], position.bikes[1], *p1territory, *p2territory);
        if (!shared_terr && COMBINED_SEARCH)
        {
            // do end game search
            p1territory = Territory::free(p1territory);
            p2territory = Territory::free(p2territory);
            Bound mybound = UPPER;
            int myscore = endgame->search(position, 0, depth, alpha, mybound);
            if (myscore == ABORT)
                return ABORT;
            if (mybound == EXACT)
                bound = EXACT;
            int opscore = endgame->search(position, 1, depth, -beta, bound);
            if (opscore == ABORT)
                return ABORT;
            score = myscore - opscore;
            if (mybound != EXACT)
            {
                if (bound == EXACT)
                {
                    if (score > 0)
                        score = WIN - (opscore / 10);
                    bound = LOWER;
                } else {
                    bound = EXACT;
                }
            } else {
                if (bound == EXACT)
                {
                    if (score > 0)
                        score = WIN - (opscore / 10);
                    else if (score < 0)
                        score = -WIN + (myscore / 10);
                } else {
                    if (score < 0)
                        score = -WIN + (myscore / 10);
                    bound = UPPER;
                }
            }
            full_search = false;
        }
        else if (depth < 1)
        {
            int score = eval(position, p1territory, p2territory);
            p1territory = Territory::free(p1territory);
            p2territory = Territory::free(p2territory);
            return score;
        } else {
            p1territory = Territory::free(p1territory);
            p2territory = Territory::free(p2territory);
        }
    }
    if (full_search)
    {
        int new_order = 0;
        for (int i=0; i < 4; i++)
        {
            int direction = (move_order >> (i*4)) & 0xF;
            position.doStep(direction);
            int new_score = -alphabeta(position, -beta, -alpha, depth-1);
            position.undoStep(direction);
            if (new_score == -ABORT)
                return ABORT;
            if (new_score > MIN_WIN)
                --new_score;
            else if (new_score < -MIN_WIN)
                ++new_score;
            if (new_score > score)
            {
                score = new_score;
                new_order = (new_order << 4) | direction;
                if (new_score > alpha)
                {
                    alpha = new_score;
                    bound = EXACT;
                    if (new_score >= beta)
                    {
                        bound = UPPER;
                        new_order |= move_order & (0xFFFF << ((i + 1) * 4));
                        break;
                    }
                }
            } else {
                new_order |= direction << (i * 4);
            }
        }
#ifdef CHECK_ORDER
        int moves = 0;
        for (int i=0; i < 4; i++)
        {
            int direction = (new_order >> (i*4)) & 0xF;
            if ((1 << direction) & moves)
            {
                fprintf(stderr, "Same direction in move order twice");
                exit(1);
            }
            moves |= 1 << direction;
        }
        if (moves != 0xF)
        {
            fprintf(stderr, "Illegal directions in move order");
            exit(1);
        }
#endif
        move_order = new_order;
    }
    ttable->set(position, depth, score, bound, move_order);
    if (clock->nodes_visited > clock->next_check)
    {
        if (clock->check_abort())
            return ABORT;
    }
    return score;
}

int ABSearch::search(Map& position)
{
    int cur_depth = -1;
    int root_scores[4];
    int root_direction[4];
    for (int i=0; i < 4; i++)
    {
        root_scores[i] = 0;
        root_direction[i] = i;
    }
    ttable->overwrite = 0;
    bool abort = false;
    bool last_search = false;
    while (!abort)
    {
        int cur_score = -WIN;
        cur_depth += 2;
        for (int i=0; i < 4; i++)
        {
            position.doStep(root_direction[i]);
            int new_score = -alphabeta(position, cur_score, WIN, cur_depth);
            position.undoStep(root_direction[i]);
            if (new_score == -ABORT)
            {
                abort = true;
                break;
            }
            root_scores[i] = new_score;
            if (new_score > cur_score)
            {
                cur_score = new_score;
            }
        }
        for (int i=1; i < 4; i++)
        {
            int score = root_scores[i];
            int direction = root_direction[i];
            int j = i;
            while (j > 0 && score > root_scores[j-1])
            {
                root_scores[j] = root_scores[j-1];
                root_direction[j] = root_direction[j-1];
                --j;
            }
            root_scores[j] = score;
            root_direction[j] = direction;
        }
        if (root_scores[0] > MIN_WIN || root_scores[0] < -MIN_WIN)
        {
            if (last_search)
                break;
            else
                last_search = true;
        } else {
            last_search = false;
        }
    }
    ttable->ageTable();
#ifdef SEARCH_REPORT
    fprintf(stderr, "\ntime %lld depth %d nodes %d score %d dir %d\t\t\t\n",
            clock->elapsed(), (cur_depth+1)/2, clock->nodes_visited,
            root_scores[0], root_direction[0]);
    fprintf(stderr, "tt overwrite %d\t\t\n", ttable->overwrite);
    fprintf(stderr, "dir   %7d %7d %7d %7d\t\t\t\n",
            root_direction[0], root_direction[1], root_direction[2],
            root_direction[3]);
    fprintf(stderr, "score %7d %7d %7d %7d\t\t\t\n",
            root_scores[0], root_scores[1], root_scores[2], root_scores[3]);
    fflush(stderr);
#endif
    return root_direction[0];
}
