
#include "endgame.h"
#include "territory.h"
#include "alphabeta.h"
#include "connectivity.h"
#include "time_util.h"

#define TT_SIZE 100

#define ABORT 110000

const char* bstr(bool b)
{
    return b ? "true" : "false";
}


EndSearch::EndSearch(StopWatch* sw, NodeTable* t, cnct::TreeSizeCache* tsc)
{
    ttable = t;
    tscache = tsc;
    clock = sw;
}

int EndSearch::dfs(Map& position, int depth, int low_score, int high_score,
        Bound& bound)
{
    ++clock->nodes_visited;
    if (position.isWall(position.bikes[0]))
    {
        bound = EXACT;
        return 0;
    }
    int move_order = 0x3210;
    const Node* node = ttable->get(position);
    if (node->hash == position.hash)
    {
        ++tthit;
        if (node->bound == EXACT
                || node->depth >= depth
                || node->score < low_score)
        {
            ++ttscore;
            bound = node->bound;
            return node->score;
        }
        move_order = node->move_order;
    }
    int treescore = tscache->get_size(*position.space, position.bikes[0],
            bound) * 10;
    int score = treescore;
    if (depth < 1 || score < low_score || bound == EXACT)
    {
        return score;
    }
    score = 0;
    bound = EXACT;
    int new_order = 0;
    low_score -= 10;
    high_score = high_score > treescore - 10 ? treescore - 10 : high_score;
    for (int i=0; i < 4; i++)
    {
        int direction = (move_order >> (i*4)) & 0xF;
        Bound child_bound = EXACT;
        position.move_bike(direction);
        int new_score = dfs(position, depth-1, low_score, high_score,
                child_bound);
        position.reset_bike(direction);
        if (new_score == ABORT)
            return ABORT;
        if (new_score > score)
        {
            score = new_score;
            bound = child_bound;
            new_order = (new_order << 4) | direction;
            if (score > low_score)
            {
                low_score = score;
            }
            if (score >= high_score)
            {
                new_order |= move_order & (0xFFFF << ((i + 1) * 4));
#ifdef SEARCH_REPORT
                if (score > high_score)
                {
                    fprintf(stderr, "\n\n\n\n\n\n\n\n\n\n\nScore higher than upper bound %d>%d\n",
                            score, high_score);
                    std::string board = position.toString();
                    fprintf(stderr, "Map:\n%s", board.c_str());
                    fprintf(stderr, "treescore %d\n", treescore);
                    exit(1);
                    score = high_score;
                }
#endif
                break;
            }
        } else {
            new_order |= direction << (i * 4);
        }
    }
    score += 10;
#ifdef CHECK_ORDER
        int moves = 0;
        for (int i=0; i < 4; i++)
        {
            int direction = (new_order >> (i*4)) & 0xF;
            if ((1 << direction) & moves)
            {
                fprintf(stderr, "Same direction in move order twice");
                exit(1);
            }
            moves |= 1 << direction;
        }
        if (moves != 0xF)
        {
            fprintf(stderr, "Illegal directions in move order");
            exit(1);
        }
#endif
    ttable->set(position, depth, score, bound, new_order);
    if (clock->nodes_visited > clock->next_check)
    {
        if (clock->check_abort())
            return ABORT;
    }
    return score;
}

int EndSearch::search(Map& start_pos, int side, int depth,
        int low_score, Bound& bound)
{
    Map position = Map(start_pos);
    if (side == 1)
    {
        position.bikes[0][0] = position.bikes[1][0];
        position.bikes[0][1] = position.bikes[1][1];
    }
    Territory* myarea = Territory::allocate();
    myarea->fill_from(*position.space, position.bikes[0]);
    (*position.space) = (*myarea);
    myarea = Territory::free(myarea);
    position.bikes[1][0] = 0;
    position.bikes[1][1] = 0;
    position.moving = 0;
    position.generateHash();
    int high_score = position.space->size(position.bikes[0]) * 10;
    return dfs(position, depth, low_score, high_score, bound);
}

int EndSearch::search(Map& start_pos)
{
    Map position = Map(start_pos);
    Territory* myarea = Territory::allocate();
    myarea->fill_from(*position.space, position.bikes[0]);
    (*position.space) = (*myarea);
    myarea = Territory::free(myarea);
    position.bikes[1][0] = 0;
    position.bikes[1][1] = 0;
    position.moving = 0;
    position.generateHash();
    int high_score = position.space->size(position.bikes[0]) * 10;
    int cur_depth = 0;
    bool root_exact[4];
    int root_scores[4];
    int root_direction[4];
    for (int i=0; i < 4; i++)
    {
        root_exact[i] = false;
        root_scores[i] = 0;
        root_direction[i] = i;
    }
    ttable->overwrite = 0;
    tthit = 0;
    ttscore = 0;
    bool abort = false;
    bool all_exact = false;
    while (!abort && !all_exact)
    {
        ++cur_depth;
        bool check_exact = true;
        for (int i=0; i < 4; i++)
        {
            if (root_exact[i])
                continue;
            Bound child_bound = EXACT;
            position.move_bike(root_direction[i]);
            int new_score = dfs(position, cur_depth, 0, high_score,
                    child_bound);
            position.reset_bike(root_direction[i]);
            if (new_score == ABORT)
            {
                abort = true;
                break;
            }
            root_scores[i] = new_score;
            root_exact[i] = child_bound == EXACT;
            if (child_bound != EXACT)
                check_exact = false;
        }
        if (!abort && check_exact)
            all_exact = true;
        for (int i=1; i < 4; i++)
        {
            bool exact = root_exact[i];
            int score = root_scores[i];
            int direction = root_direction[i];
            int j = i;
            while (j > 0 && score > root_scores[j-1])
            {
                root_exact[j] = root_exact[j-1];
                root_scores[j] = root_scores[j-1];
                root_direction[j] = root_direction[j-1];
                --j;
            }
            root_exact[j] = exact;
            root_scores[j] = score;
            root_direction[j] = direction;
        }
    }
    ttable->ageTable();
#ifdef SEARCH_REPORT
    fprintf(stderr, "\ntime %lld depth %d nodes %d\t\t\n",
            clock->elapsed(), cur_depth, clock->nodes_visited);
    fprintf(stderr, "tthit %d ttscore %d overwrite %d\t\t\n",
            tthit, ttscore, ttable->overwrite);
    fprintf(stderr, "dir   %7d %7d %7d %7d\t\t\t\n",
            root_direction[0], root_direction[1], root_direction[2],
            root_direction[3]);
    fprintf(stderr, "score %7d %7d %7d %7d\t\t\t\n",
            root_scores[0], root_scores[1], root_scores[2], root_scores[3]);
    fprintf(stderr, "exact %7s %7s %7s %7s\n",
            bstr(root_exact[0]), bstr(root_exact[1]),
            bstr(root_exact[2]), bstr(root_exact[3]));
    if (all_exact)
        fprintf(stderr, "exact result\n");
    else
        fprintf(stderr, "\t\t\t\n");
#endif
    return root_direction[0];
}

